/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package TexturedSphere;

// General Imports
import java.awt.*;
import java.io.*;
import java.util.*;

// JOGL Imports
import javax.media.opengl.*;
import javax.media.opengl.glu.GLUquadric;
import javax.media.opengl.glu.GLU;
import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLEventListener;

/**
 *
 * @author mzaczek
 */
public class Renderer implements GLEventListener {    
    /** Scale Factor for zooming */
    private float scaleAll = 1.0f;

    /** Rotation factor around X */
    private float rotX = 0.0f;

    /** Rotation factor around Y */        
    private float rotY = 0.0f;

    /** A mouse point */
    private Point mousePoint = new Point();

    /** The closest that one can zoom in to */
    private static final float MIN_SCALE = 0.1f;

    /** The farthest that one can zoom out to */
    private static final float MAX_SCALE = 10000.9f;

    /** The radius of the boundary of the model that is loaded */
    private float radius = 1.0f;

    private float lightPosition[] = { -500.0f, 500.0f, 0.0f, 1.0f };        

    /** Ambient light array */
    private float lightAmbient[] = {0.4f, 0.4f, 0.4f, 0.3f};

    /** Diffuse light array */
    private float lightDiffuse[] = {0.3f, 0.3f, 0.3f, 1.0f};             

    /** A 3D Sphere to be textured */
    private GLUquadric texturedSphere;
    private GLUquadric wireframeSphere;

    /** Name of the texture to load for the star background */
    private String textureName = new String("src/TexturedSphere/Earth.jpg");

    /** Array of texture ids */
    private int textures[] = new int[2];        

    /** GLU object */
    private GLU glu = new GLU();

    /** Light Type false=point, true=infinite */
    private boolean lightType = true;
    
    /** Shade model: false=FLAT, true=SMOOTH */
    private boolean shadeModel = true;
    
    /** If textures are enabled */
    private boolean texturesEnabled = true;
    
    private int tess = 32;
    private boolean wireframe = true;
    
    /**
     * Display method
     * 
     * @param gLDrawable
     */
    public void display(GLAutoDrawable gLDrawable) {
        final GL gl = gLDrawable.getGL();

        // Clear The Screen And The Depth Buffer
        gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);

        // Reset The View
        gl.glLoadIdentity();

        // Enable lighting
        gl.glEnable(GL.GL_LIGHTING);
        
        if (texturesEnabled) {
            // Enable the textures
            gl.glEnable(GL.GL_TEXTURE_2D);

            // Select the texture image
            gl.glBindTexture(GL.GL_TEXTURE_2D, textures[0]);
        }

        // Update the light location
        gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, lightPosition, 0);    

        // Update the shade model: FLAT or SMOOTH
        if (shadeModel) {
            gl.glShadeModel(GL.GL_SMOOTH);
        } else {
            gl.glShadeModel(GL.GL_FLAT);					
        }

        // Make sure the rotations don't keep growing
        if (Math.abs(rotX) >= 360.0f) rotX = 0.0f;
        if (Math.abs(rotY) >= 360.0f) rotY = 0.0f;              

        // Draw the scene (by default, the lighting, material and textures 
        // are enabled/disabled within the renderer for the model)
        gl.glPushMatrix();
            // Scale the model (used for zooming purposes)
            gl.glScalef(scaleAll, scaleAll, scaleAll);

            // Rotate the model based on mouse inputs
            gl.glRotatef(rotY, 1.0f, 0.0f, 0.0f);
            gl.glRotatef(rotX, 0.0f, 0.0f, 1.0f);

            // Draw the textured sphere
            glu.gluSphere(texturedSphere, radius, tess, tess);

            if (wireframe) {
                // Draw the wireframe sphere
                gl.glDisable(GL.GL_LIGHTING);
                gl.glColor3f(1.0f, 0.9f, 0.7f);
                glu.gluSphere(wireframeSphere, radius*1.004f, tess, tess);   
                gl.glColor3f(1.0f, 1.0f, 1.0f);
            }
        gl.glPopMatrix();

        // Disable the textures
        gl.glDisable(GL.GL_TEXTURE_2D);  
        
        // Disable lighting
        gl.glDisable(GL.GL_LIGHTING);

        // Flush the draw buffer
        gl.glFlush();
    }


    /** Called when the display mode has been changed.  <B>!! CURRENTLY UNIMPLEMENTED IN JOGL !!</B>
     * @param gLDrawable The GLDrawable object.
     * @param modeChanged Indicates if the video mode has changed.
     * @param deviceChanged Indicates if the video device has changed.
     */
    public void displayChanged(GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged) {}

    /** Called by the drawable immediately after the OpenGL context is
     * initialized for the first time. Can be used to perform one-time OpenGL
     * initialization such as setup of lights and display lists.
     * @param gLDrawable The GLDrawable object.
     */
    public void init(GLAutoDrawable gLDrawable) {
        final GL gl = gLDrawable.getGL();     

        // Load the textures
        try {
            loadGLTextures(gl);	
        } catch (IOException e) {
            throw new RuntimeException(e);
        }            

        // Create Pointers To The Quadric Objects (Return 0 If No Memory)
        texturedSphere = glu.gluNewQuadric();
        wireframeSphere = glu.gluNewQuadric();          

        // Create Smooth Normals
        glu.gluQuadricNormals(texturedSphere, GLU.GLU_SMOOTH);
        glu.gluQuadricNormals(wireframeSphere, GLU.GLU_SMOOTH);

        // Set the textured sphere to be filled and the wireframe sphere to be wired
        glu.gluQuadricDrawStyle(texturedSphere, GLU.GLU_FILL);
        glu.gluQuadricDrawStyle(wireframeSphere, GLU.GLU_LINE);

        // Create Texture Coords for the textured sphere
        glu.gluQuadricTexture(texturedSphere, true);

        // Set The Texture Generation Mode For S To Sphere Mapping
        gl.glTexGeni(GL.GL_S, GL.GL_TEXTURE_GEN_MODE, GL.GL_SPHERE_MAP);

        // Set The Texture Generation Mode For T To Sphere Mapping (NEW) 
        gl.glTexGeni(GL.GL_T, GL.GL_TEXTURE_GEN_MODE, GL.GL_SPHERE_MAP);
        
        // Setup the shading model (FLAT or SMOOTH)
        if (shadeModel) {
            gl.glShadeModel(GL.GL_SMOOTH);
        } else {
            gl.glShadeModel(GL.GL_FLAT);					
        }
        
        // Setup the specular light to use the diffuse light properties
        gl.glMaterialfv(gl.GL_FRONT, gl.GL_SPECULAR, lightDiffuse, 0);
        
        // Setup the ambient and diffuse light
        gl.glMaterialfv(gl.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE, lightAmbient, 0);

        // Background - black
        gl.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);			
        
        // Depth Buffer Setup
        gl.glClearDepth(1.0f);	
        
        // Clear The Stencil Buffer To 0
        gl.glClearStencil(0);	
        
        // Enables Depth Testing
        gl.glEnable(GL.GL_DEPTH_TEST);					

        // Really Nice Perspective Calculations
        gl.glHint(GL.GL_PERSPECTIVE_CORRECTION_HINT, GL.GL_NICEST);	
        
        // Set The Ambient Lighting For Light0
        gl.glLightfv(GL.GL_LIGHT0, GL.GL_AMBIENT, lightAmbient, 0);	
        
        // Set The Diffuse Lighting For Light0
        gl.glLightfv(GL.GL_LIGHT0, GL.GL_DIFFUSE, lightDiffuse, 0);	
        
        // Set The Position For Light0
        gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, lightPosition, 0);	
        
        // ?
        gl.glLightModeli(GL.GL_LIGHT_MODEL_LOCAL_VIEWER, GL.GL_FALSE);
        
        // ?
        gl.glLightModelfv(gl.GL_LIGHT_MODEL_AMBIENT, lightAmbient, 0);    
        
        // Enable Lighting
        gl.glEnable(GL.GL_LIGHTING);					
        
        // Enable Light 0
        gl.glEnable(GL.GL_LIGHT0);
        
        // Enable face culling
        gl.glEnable(GL.GL_CULL_FACE);
        
        // Enable normalization 
        gl.glEnable(GL.GL_NORMALIZE);
        
        // ?
        gl.glLightModelf(GL.GL_LIGHT_MODEL_COLOR_CONTROL,GL.GL_SEPARATE_SPECULAR_COLOR);             
    }


    /** Called by the drawable during the first repaint after the component has
     * been resized. The client can update the viewport and view volume of the
     * window appropriately, for example by a call to
     * GL.glViewport(int, int, int, int); note that for convenience the component
     * has already called GL.glViewport(int, int, int, int)(x, y, width, height)
     * when this method is called, so the client may not have to do anything in
     * this method.
     * @param gLDrawable The GLDrawable object.
     * @param x The X Coordinate of the viewport rectangle.
     * @param y The Y coordinate of the viewport rectanble.
     * @param width The new width of the window.
     * @param height The new height of the window.
     */
    public void reshape(GLAutoDrawable gLDrawable, int x, int y, int width, int height)
    {
        final GL gl = gLDrawable.getGL();
        final GLU glu = new GLU();

        if (height <= 0) // avoid a divide by zero error!
            height = 1;
        final float h = (float)width / (float)height;
        gl.glViewport(0, 0, width, height);
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glLoadIdentity();
        gl.glOrtho(-radius, radius, -radius, radius, -10*radius, 10*radius);
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glLoadIdentity();
    }

    /**
     * Load textures for the Stars and Sun
     *
     * @param gl The GL instance
     * @throws IOException
     */
    private void loadGLTextures( GL gl ) throws IOException {
        gl.glGenTextures(2, textures, 0);

        TextureReader.Texture texture = TextureReader.readTexture(textureName);
        gl.glBindTexture(GL.GL_TEXTURE_2D, textures[0]);
        gl.glTexImage2D(GL.GL_TEXTURE_2D,
                        0,
                        GL.GL_RGB8,
                        texture.getWidth(),
                        texture.getHeight(), 0, GL.GL_RGB, GL.GL_UNSIGNED_BYTE, texture.getPixels());
        gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_REPEAT);
        gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_REPEAT);
        gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR);
        gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR);
    }        

    /** 
     * Get the point at the start of the mouse drag
     * 
     * @param MousePt
     */
    void startDrag( Point MousePt ) {
        mousePoint.x = MousePt.x;
        mousePoint.y = MousePt.y;  
    }       

    /**
     * Calculate the delta and rotation values for the dragging of the mouse
     * @param MousePt
     */
    void drag( Point MousePt ) {
        Point delta = new Point();
        delta.x = MousePt.x - mousePoint.x;
        delta.y = MousePt.y - mousePoint.y;

        mousePoint.x = MousePt.x;
        mousePoint.y = MousePt.y;

        rotX += delta.x * 2.0f / scaleAll;
        rotY += delta.y * 2.0f / scaleAll;
    }

    /** 
     * Get the point at the start of the mouse zoom action
     * @param MousePt
     */
    void startZoom( Point MousePt ) {
        mousePoint.x = MousePt.x;
        mousePoint.y = MousePt.y;
    }

    /** 
     * Caclaulte the scaling parameters for zooming while the zoom drag
     * is ongoing
     * 
     * @param MousePt
     */
    void zoom( Point MousePt ) {       
        Point delta = new Point();
        delta.x = MousePt.x - mousePoint.x;
        delta.y = MousePt.y - mousePoint.y;

        mousePoint.x = MousePt.x;
        mousePoint.y = MousePt.y;

        float addition = (delta.x + delta.y) / 2.0f;

        if (addition < 0.0  &&  (scaleAll+addition) > MIN_SCALE) {
            scaleAll += addition;            
        } 

        if (addition > 0.0  &&  (scaleAll+addition) < MAX_SCALE) {
            scaleAll += addition;
        }           
    }
        
        
    public void moveLightXYZ( int direction, float distance ) {
        //System.out.println("Moving light: " + direction + " by " + distance);
        if (direction == 1) {
            lightPosition[0] += distance;
        } else if (direction == 2) {
            lightPosition[1] += distance;
        } else if (direction == 3) {
            lightPosition[2] += distance;         
        }
        System.out.println("Light: " + lightPosition[0] + ", " + lightPosition[1] + ", " + lightPosition[2]);        
    }
    
  
    public void scale( float val ) {
        lightPosition[0] *= val;
        lightPosition[1] *= val;
        lightPosition[2] *= val;
        System.out.println("Light: " + lightPosition[0] + ", " + lightPosition[1] + ", " + lightPosition[2]);        
    }
    
    
    public void setLightType( boolean type ) {
        lightType = type;
        
        if (lightType) {
            lightPosition[3] = 0.0f;
        } else {
            lightPosition[3] = 1.0f;
        }
    }
        
    public boolean getLightType() {
        return lightType;
    }   
    
    public void setShadeModel( boolean type ) {
        shadeModel = type;
    }
        
    public boolean getShadeModel() {
        return shadeModel;
    }
    
    public void setTextureStatus( boolean type ) {
        texturesEnabled = type;
    }
    
    public boolean getTextureStatus() {
        return texturesEnabled;
    }    
    
    public void increaseTess() {
        tess *= 2;
        
        System.out.println("Number of tesselations: " + tess);
    }
    
    public void decreaseTess() {
        tess /= 2;
        
        if (tess < 2) {
            tess = 2;
        }
        
        System.out.println("Number of tesselations: " + tess);        
    }
    
    public void toggleWireframe() {
        wireframe = !wireframe;
    }
}
